import json
import yaml
import os
from cnvrg.modules.project import Project
from cnvrg.modules.flow_version import FlowVersion
from cnvrg.helpers.url_builder_helper import url_join
from cnvrg.modules.errors import CnvrgError

import cnvrg.helpers.param_build_helper as param_build_helper
import cnvrg.helpers.apis_helper as apis_helper
import cnvrg.modules.errors as errors


class Flow():
    def __init__(self, slug, project=None, version=None):
        owner, project_slug, slug = param_build_helper.parse_params(slug, param_build_helper.FLOW)

        p = Project.factory(owner, project)

        if p is None:
            p = Project(url_join(owner, project_slug))

        self.slug = slug
        self.version = version or "latest"
        self.owner = owner
        self.project = p

    def run(self):
        project = self.project
        resp = apis_helper.post(url_join(project.get_base_url(), 'flows', 'run_flow'), data={"flow_slug": self.slug})
        status = resp["status"]
        if status == 200:
            fv_title = resp["flow_version"]["title"]
            return FlowVersion(self, fv_title, project)
        else:
            raise CnvrgError("Could not create flow")

    def __latest_artifacts(self, task=None):
        resp = apis_helper.get_v2(
            url_join(
                self.project.get_base_url(api="v2"),
                'flows',
                self.slug,
                "latest_end_commits?exec_task_title={}".format(task),
            )
        )

        if not resp:
            return None

        return resp.json().get("artifacts")

    def artifacts(self, task=None):
        artifacts = self.__latest_artifacts(task=task)
        if not artifacts: return []
        return artifacts

    def pull_artifacts(self, path=".", task=None):
        artifacts = self.artifacts(task=task)
        mapped_files = []
        for artifact in artifacts:
            try:
                fpath = os.path.join(path, artifact.get("fullpath"))
                os.makedirs(os.path.dirname(fpath), exist_ok=True)
                artifact_path = apis_helper.download_file(artifact.get("url"), fpath)
                mapped_files.append({**artifact, **{"path": os.path.abspath(artifact_path)}})
            except Exception as e:
                print(e)
                print("Error while downloading {}".format(artifact.get("name")))
        return mapped_files

    @staticmethod
    def create(file=None, yaml_content=None, project=None):
        if not file and not yaml_content:
            raise errors.CnvrgError("File or yaml is missing")

        project = project or Project()

        if yaml_content:
            try:
                data = yaml.safe_load(yaml_content)
            except Exception:
                raise errors.CnvrgError("Please check that your yaml valid")
        else:
            try:
                with open(file, 'r') as f:
                    data = yaml.safe_load(f)
            except Exception:
                raise errors.CnvrgError("Please check that your yaml exists")

        if not data:
            raise errors.CnvrgError("Yaml can't be empty")

        resp = apis_helper.post(url_join(project.get_base_url(), 'flows'), data={"flow_version": json.dumps(data)})

        status = resp["status"]
        if status == 200:
            flow_title = resp["flow_version"]["title"]
            print("Flow {} created successfully".format(flow_title))
            flow_slug = resp["flow_version"]["flow_id"]
            return Flow(flow_slug, project)
        else:
            raise CnvrgError("Could not create flow")




from cnvrg.modules.cnvrg_job import CnvrgJob
from cnvrg.modules.project import Project
import cnvrg.helpers.string_helper as string_helper
import cnvrg.helpers.param_build_helper as param_build_helper
from cnvrg.helpers.env_helper import CURRENT_JOB_TYPE, CURRENT_JOB_ID, ENDPOINT, POOL_SIZE
from cnvrg.modules.errors import UserError
from cnvrg.helpers.url_builder_helper import url_join
from cnvrg.helpers.error_catcher import suppress_exception
import cnvrg.helpers.parallel_helper as parallel_helper
import cnvrg.helpers.apis_helper as apis_helper
from cnvrg.modules.dataset import Dataset
from cnvrg.modules.errors import CnvrgError
import json
import sys
import time
import os
import requests
import datetime
import traceback
import pandas as pd
import csv

from cnvrg.modules.endpoint_version_data import EndpointVersionData


class Endpoint(CnvrgJob):
    def __init__(self, endpoint=None, project=None):
        owner, project_slug, slug = param_build_helper.parse_params(endpoint, param_build_helper.ENDPOINT)

        p = Project.factory(owner, project)
        if p is None:
            p = Project(url_join(owner, project_slug))

        if not slug:
            raise UserError("Cant create an endpoint without slug")
        slug = slug or CURRENT_JOB_ID
        self.model_name = os.environ.get("CNVRG_MODEL_NAME", "unknown")
        self.log_input = self.should_log_input()
        self.log_output = self.should_log_output()

        self.project = p
        super(Endpoint, self).__init__(slug, ENDPOINT, p)
        self.__data = None

    def __base_log(self, **params):
        log = {
            "job_id": self.job_slug,
            "job_type": self.job_type,
            "owner": self.job_owner_slug,
            "project": self.job_project_slug,
            "model": self.model_name,
            "event_time": time.time(),
            **params
        }
        print(json.dumps(log))
        sys.stdout.flush()

    def log_request(self, input_params="", result="", start_time=None):
        hash = {}
        if self.log_input:
            input_str = str(input_params)
            if len(input_str) > 1000:
                print("Input was cropped due to log size (limited to 1000 chars)")
                print("full input: {}".format(input_str))
                input_str = input_str[:1000]
            hash["input"] = input_str
        else:
            hash["input"] = ""
        if self.log_output:
            output_str = str(result)
            if len(output_str) > 1000:
                print("Output was cropped due to log size (limited to 1000 chars)")
                print("full output: {}".format(output_str))
                output_str = output_str[:1000]
            hash["output"] = output_str
        else:
            hash["output"] = ""
        hash["start_time"] = str(datetime.datetime.utcfromtimestamp(int(start_time)))
        end_time = time.time()
        elapsed_time = int((end_time - start_time) * 1000)  # ms
        hash["elapsed_time"] = elapsed_time
        self.__base_log(**hash)

    def log_error(self, e, trace=False):
        hash = {}
        hash["error"] = str(e)
        if trace:
            hash["traceback"] = traceback.format_exc()
        self.__base_log(**hash)

    def log_param(self, tag_name=None, tag_value=None):
        hash = {}
        hash[tag_name] = tag_value
        hash["cnvrg_tag_name"] = tag_name
        self.__base_log(**hash)

    def log_metric(self, name=None, y:float=None, x=None):
        hash = {}
        try:
            hash[name] = float(y)
            hash["y"] = float(y)
        except Exception as e:
            print("ERROR {}".format(e), file=sys.stderr)
            return
        hash["x"] = x
        hash["cnvrg_metric_name"] = name
        self.__base_log(**hash)

    def log_prediction(self, input=None, output=None, elapsed_time=0):
        start_time = datetime.datetime.utcnow().strftime("%c")
        self.__base_log(input=input, output=output, elapsed_time=elapsed_time, start_time=start_time)

    def scale_up(self):
        return apis_helper.post_v2(
            url_join(self.project.get_base_url(api="v2"), 'endpoints', self.data["slug"], "scale_up"), data={})

    def scale_down(self):
        return apis_helper.post_v2(
            url_join(self.project.get_base_url(api="v2"), 'endpoints', self.data["slug"], "scale_down"), data={})

    def update_model(self, title=None, kind=None, input_file=None, file_name=None, function_name=None,
                     kafka_brokers=None, kafka_input_topics=None, kafka_output_topics=None, templates_names=None,
                     image_slug=None, commit=None, prep_file=None, prep_function=None, git_commit=None,
                     git_branch=None, last_deployment_commits=False):
        self._refresh_data()
        if self.data["last_deployment"] == None or self.data["status"] == "stopped":
            raise CnvrgError("Can't find last deployment")
        if self.data["deployments"][-1]["is_update"]:
            raise CnvrgError("Can't update an endpoint that is currently updating")
        if templates_names and not isinstance(templates_names, list):
            raise CnvrgError("templates_names must be in a list")
        if kafka_brokers and not isinstance(kafka_brokers, list):
            raise CnvrgError("kafka_brokers must be in a list")
        if kafka_input_topics and not isinstance(kafka_input_topics, list):
            raise CnvrgError("kafka_input_topics must be in a list")
        if kafka_output_topics and not isinstance(kafka_output_topics, list):
            raise CnvrgError("kafka_output_topics must be in a list")

        if last_deployment_commits:
            git_commit = self.data["last_deployment"]["git_commit"]
            commit = self.data["last_deployment"]["commit"]

        old_kind = self.data["kind"]
        if kind is None:
            if old_kind == "webserver":
                new_kind = 0
            elif old_kind == "kafka_stream":
                new_kind = 1
            elif old_kind == "batch":
                new_kind = 2
        else:
            new_kind = kind

        data = {
            "model": {
                "title": title or int(self.data["last_deployment"]["title"]) + 1,
                "kind": new_kind,
                "input_file": input_file or self.data["last_deployment"]["input_file"],
                "file_name": file_name or self.data["last_deployment"]["file_name"],
                "function_name": function_name or self.data["last_deployment"]["function_name"],
                "kafka_brokers": kafka_brokers or self.data["last_deployment"]["kafka_brokers"],
                "kafka_input_topics": kafka_input_topics or self.data["last_deployment"]["kafka_input_topics"],
                "kafka_output_topics": kafka_output_topics or self.data["last_deployment"]["kafka_output_topics"],
                "image_slug": image_slug or self.data["last_deployment"]["image_slug"],
                "templates_names": templates_names,
                "templates_ids": [t["id"] for t in self.data["templates"]],
                "from_deployment_slug": self.data["last_deployment"]["slug"],
                "env_setup": self.data["last_deployment"]["env_setup"],
                "commit": commit or None,
                "prep_file": prep_file or self.data["last_deployment"]["prep_file"],
                "prep_function": prep_function or self.data["last_deployment"]["prep_function"],
                "git_commit": git_commit or "latest",
                "git_branch": git_branch or self.data["last_deployment"]["git_branch"]
            }
        }
        return apis_helper.post_v2(
            url_join(self.project.get_base_url(api="v2"), 'endpoints', self.data["slug"], "update_models"), data)

    def link_experiment(self):
        job_id = os.environ.get("CNVRG_JOB_ID", "unknown")
        job_type = os.environ.get("CNVRG_JOB_TYPE", "unknown")
        resp = None
        if job_type == "Experiment":
            resp = apis_helper.post_v2(
                url_join(self.project.get_base_url(api="v2"), 'endpoints', self.data["slug"], "link_experiment"),
                data={"experiment_id": job_id})
        return resp

    def is_deployment_running(self):
        resp = apis_helper.get_v2(
            url_join(self.project.get_base_url(api="v2"), 'endpoints', self.data["slug"], "is_running"))
        return resp.json().get("status")

    def rollback(self):
        resp = apis_helper.post_v2(
            url_join(self.project.get_base_url(api="v2"), 'endpoints', self.data["slug"], "rollback_model"),
            data={})
        return resp.json()

    def get_predictions(self, model_title=None):
        url = url_join(self.project.get_base_url(api="v2"), 'endpoints', self.data["slug"], "predictions")
        if model_title:
            url = url_join(url, "?model_title={model_title}".format(model_title=model_title))
        resp = apis_helper.get_v2(url)
        return resp.json()

    def run_batch_predict(self, dataset, input_file, output_file, scale: bool = True):
        if not output_file or not input_file or not dataset:
            raise CnvrgError("output_file, input_file and dataset can\'t be empty")

        try:
            endpoint_type = self.data.get("kind")
            should_scale = endpoint_type == "batch" and scale

            # fetch dataset details
            ds = Dataset(dataset)
            if ds is None:
                raise CnvrgError("Can\'t find Dataset {dataset}".format(dataset=dataset))

            ds_url = ds.get_full_url()

            self.link_experiment()

            if should_scale:
                print("Starting to scale up endpoint")
                self.scale_up()

                is_running = self.is_deployment_running()
                while not is_running:
                    print("Endpoint is not running yet, retrying in 10 seconds")
                    time.sleep(10)
                    is_running = self.is_deployment_running()
                print("Endpoint is online, starting batch prediction")

                time.sleep(20)
            else:
                is_running = self.is_deployment_running()
                if is_running:
                    print("Starting batch prediction")
                else:
                    raise CnvrgError("Endpoint is not running please send scale: True or start your endpoint")

            # Input file should be absolute path
            row_list = []
            data = pd.read_csv(input_file, header=0)
            if not data:
                raise CnvrgError("Input file is empty or invalid")
            for row in data.values:
                try:
                    r_list = row.tolist()
                    resp = self.predict(r_list)
                    row_list.append([r_list, resp.get("prediction")])
                except Exception as e:
                    raise CnvrgError(e)

            # create output file tree if not exists
            dirname = os.path.dirname(output_file)
            if dirname:
                os.makedirs(dirname, exist_ok=True)

            # Output file should be absolute path in /cnvrg
            with open(output_file, 'w') as file:
                writer = csv.writer(file)
                writer.writerow(["input", "prediction"])
                for row in row_list:
                    writer.writerow(row)

            print('Uploading {output_file} to dataset {dataset}'.format(output_file=output_file, dataset=dataset))
            os.system('cnvrg data put {url} {exported_file}'.format(url=ds_url, exported_file=output_file))

            print("Batch prediction has finished")
            if should_scale:
                print("Scaling down endpoint")
                self.scale_down()
        except Exception as e:
            if should_scale:
                self.scale_down()
            raise e

    @property
    def session(self):
        return self.__session()

    @property
    def data(self):
        if self.__data:
            return self.__data
        self._refresh_data()
        return self.__data

    @property
    def title(self):
        return self.data.get("title")

    @property
    def url(self):
        return self.data.get("endpoint_url")

    @property
    def model_url(self):
        return self.data.get("public_url")

    @property
    def model_grpc_url(self):
        return self.data.get("grpc_url")

    @property
    def min_replica(self):
        return self.data.get("min_replica")

    @property
    def max_replica(self):
        return self.data.get("max_replica")

    @property
    def active_versions(self):
        return len(self.data.get("deployments"))

    @property
    def status(self):
        return self.data.get("status")

    @suppress_exception
    def __get_endpoint(self):
        return apis_helper.get(self.__base_url()).get("endpoint")

    def __base_url(self):
        return url_join(
            self.project.get_base_url(), string_helper.to_snake_case(self.job_type) + "s", self.job_slug
        )

    def _refresh_data(self):
        self.__data = self.__get_endpoint()

    def __session(self):
        session = requests.session()
        session.headers = {
            **apis_helper.JSON_HEADERS,
            "Cnvrg-Api-Key": self.data.get("api_key")
        }
        return session

    def version(self, version):
        if 0 <= version < len(self.data.get("deployments")):
            return EndpointVersionData(self.data.get("deployments")[version])
        return EndpointVersionData(self.data.get("last_deployment"))

    def predict(self, o):
        endpoint_url = self.data.get("endpoint_url")
        if not endpoint_url:
            self._refresh_data()
            endpoint_url = self.data.get("endpoint_url")
        return self.session.post(endpoint_url, data=json.dumps({"input_params": o})).json()

    def single_batch_predict(self, o):
        prediction = self.predict(o)
        if isinstance(o, dict):
            return {**o, **prediction}
        else:
            return {"query": o, **prediction}

    def disable_log_input(self):
        self.log_input = False
        os.environ["CNVRG_ENDPOINT_LOG_INPUT"] = "false"

    def should_log_input(self):
        return os.environ.get("CNVRG_ENDPOINT_LOG_INPUT", "").lower() != "false"

    def disable_log_output(self):
        self.log_output = False
        os.environ["CNVRG_ENDPOINT_LOG_OUTPUT"] = "false"

    def should_log_output(self):
        return os.environ.get("CNVRG_ENDPOINT_LOG_OUTPUT", "").lower() != "false"



import cnvrg.helpers.apis_helper as apis_helper
from cnvrg.helpers.url_builder_helper import url_join
from cnvrg.helpers.param_helper import wrap_into_list
from cnvrg.helpers import logger_helper
from typing import AnyStr, List

class ChartsMixin(object):

    ### Verifiers
    def __verify_same_type(self, payload):
        if len(set([o.__class__ for o in payload])) > 1:
            raise ValueError("Got two charts from different type.")

    def __inflate_y_ticks(self, data, y_ticks):
        if y_ticks: return y_ticks
        for series in data:
            if series.y_ticks: return series.y_ticks

    def __inflate_x_ticks(self, data, x_ticks):
        if x_ticks: return x_ticks
        for series in data:
            if series.x_ticks: return series.x_ticks

    def __only_one_heatmap(self, data: List):
        heatmaps = len(list(filter(lambda o: o.chart_type() == 'heatmap', data)))
        if heatmaps > 1:
            raise ValueError("Can't generate chart with more than a single heatmap")

    def __add_series_name(self, data):
        for idx, series in enumerate(data):
            series.set_series_index(idx)

    def log_chart(self, key: AnyStr, data, group=None, step=None, title: AnyStr=None,
                  x_axis: AnyStr=None, y_axis: AnyStr=None, y_ticks: List=None, x_ticks: List=None):
        data = wrap_into_list(data)
        self.__verify_same_type(data)
        self.__only_one_heatmap(data)
        self.__add_series_name(data)
        y_ticks = self.__inflate_y_ticks(data, y_ticks)
        x_ticks = self.__inflate_x_ticks(data, x_ticks)

        additional_params = {}

        if data[0].chart_type() == 'heatmap':
            chart = data[0]
            if chart.color_stops:
                additional_params["stops"] = chart.color_stops

        if data[0].chart_type() == 'bar':
            chart = data[0]
            if chart.colors:
                additional_params["colors"] = chart.colors

        if data[0].chart_type() == 'heatmap' or data[0].chart_type() == 'bar':
            chart = data[0]
            if chart.max:
                additional_params["max"] = chart.max
            if chart.min:
                additional_params["min"] = chart.min

        serieses = [o.as_series() for o in data]
        resp = self.__create_chart(
            key=key,
            group=group,
            step=step,
            chart_type=data[0].chart_type(),
            serieses=serieses,
            title=title,
            x_axis=x_axis,
            y_axis=y_axis,
            x_ticks=x_ticks,
            y_ticks=y_ticks,
            **additional_params
        )
        if resp.get("status") == 200:
            logger_helper.log_message("Chart {tag} created successfully.".format(tag=resp.get("tag")))
        else:
            logger_helper.log_message(resp.get("error"))

    def __create_chart(self, key, group, step, chart_type, serieses, title, x_axis, y_axis, x_ticks, y_ticks, **kwargs):
        data = {
            "chart": {
                "chart_type": chart_type,
                "title": title,
                "group": group,
                "step": step,
                "key": key,
                "serieses": serieses,
                "x_axis": x_axis,
                "y_axis": y_axis,
                "x_ticks": x_ticks,
                "y_ticks": y_ticks,
                **kwargs,
            }
        }
        resp = apis_helper.post(url_join(self._base_url(), 'charts'), data=data)
        return resp


import cnvrg.helpers.libs_helper as libs_helper
from keras.callbacks import Callback
from cnvrg.modules.experiment import Experiment
class CnvrgKerasCallback(Callback):
    def __init__(self):
        self.experiment = Experiment()
        super(CnvrgKerasCallback, self).__init__()

    def on_epoch_begin(self, epoch=None, logs=None):
        self.experiment.log("Train began")
        self.losses = []

    def on_epoch_end(self, epoch=None, logs=None):
        xs = []
        ys = []
        grouping = []
        for group, value in logs.items():
            if group == 'epoch': continue
            grouping.append(group)
            ys.append(value)
            xs.append(epoch)
        self.experiment.log_metric("Train", ys, xs, grouping)
        print("Sending cnvrg logs", logs)
        self.losses = []


    def on_batch_begin(self, batch, logs=None):
        self.losses.append(logs.get('loss'))


    def on_batch_end(self, batch, logs=None):
        self.losses.append(logs.get('loss'))

    def on_train_begin(self, logs=None):
        print(logs)

    def on_train_end(self, logs=None):
        print(logs)


import sys
import io
import time
import threading
import builtins as __builtin__
import re
ANSI_ESCAPE = re.compile(r'\x1B\[[0-?]*[ -/]*[@-~]')

def before_run():
    olderrio = sys.stderr
    oldstdio = sys.stdout
    sys.stdout = std_buffer = io.StringIO()
    sys.stderr = err_buffer = io.StringIO()
    return std_buffer, err_buffer, olderrio, oldstdio


def after_run(buffer, err_buffer, errio, stdio):
    buffer.close()
    err_buffer.close()
    sys.stdout = stdio
    sys.stderr = errio



def remove_colors(block):
    first_filter = [ANSI_ESCAPE.sub('', line) for line in block.strip().split("\n")]
    return [line for line in filter(lambda x: x, first_filter)]

def check_buffer_thread(buffer, callback):
    def p(seek):
        v = buffer.getvalue()
        v = v[seek:]
        if v:
            lines = remove_colors(v)
            callback(lines)
        return seek + len(v)

    seek = 0
    while not buffer.closed:
        seek = p(seek)
        time.sleep(0.5)

def get_buffer_callbacks(callback, f_io):
    callbacks = []
    callbacks.append(lambda x: print("\n".join(map(str, x)), file=f_io))
    if callback: callbacks.append(callback)
    return lambda x: [cb(x) for cb in callbacks]


def run_callable(callable, arguments: list=None, callback=None, err_callback=None):
    ### the idea here is to "redirect" to stdout/stderr to buffers.


    ### on error: it will log to err_buffer instead of sys.stderr
            ##      from err_buffer it will go to stderr and to
    buffer, err_buffer, errio, stdio = before_run()

    ## those threads will monitor the buffer and will call the relevant "prints"
    t = threading.Thread(target=check_buffer_thread, args=(buffer, get_buffer_callbacks(callback, stdio)))
    terr = threading.Thread(target=check_buffer_thread, args=(err_buffer, get_buffer_callbacks(err_callback, errio)))
    t.start()
    terr.start()
    try:
        exit_status = callable(*arguments)
    except Exception as e:
        exit_status = 1
        err_callback(str(e))
    except KeyboardInterrupt as e:
        exit_status = -100
        err_callback("Keyboard Interrupt")
    time.sleep(0.5)
    after_run(buffer, err_buffer, errio, stdio)
    t.join()
    terr.join()
    return exit_status




import sys
import inspect

class NoTraceBackWithLineNumber(Exception):
    def __init__(self, msg):
        try:
            ln = sys.exc_info()[-1].tb_lineno
        except AttributeError:
            ln = inspect.currentframe().f_back.f_lineno
        self.args = "{0.__name__}: {1}".format(type(self), msg),
        sys.exit(self)
class CnvrgError(NoTraceBackWithLineNumber):
    pass

class UserError(CnvrgError):
    pass

class NotImplementedError(CnvrgError):
    pass

class UnknownStsError(CnvrgError):
    pass

class APINotFound(Exception):
    pass


import os
import math
import json
import time
import hashlib
import fnmatch
import mimetypes
import threading

from queue import Queue
from pathlib import Path
from os import path, walk
from cnvrg.helpers.url_builder_helper import url_join
from cnvrg.modules.storage.progress import Progressbar

import cnvrg.modules.storage as storage
import cnvrg.helpers.apis_helper as apis_helper


# File type check util
def filter_by_mimetypes(filters, current_path):
    """
    mimetypes for files will return in the following format: "TYPE/EXACT_TYPE", for ex. 'image/jpeg'
    This function will receive general types in the filters list (for ex. ['image']), and will only return
    True for files that match any of the filters
    @param filters: Allowed file types.
    @param path: File to test against filters
    @return: True if file passed filters, False if file is filtered
    """
    # This means its a directory
    if not mimetypes.guess_type(current_path)[0]:
        return False
    if not filters:
        return True
    mime_type = mimetypes.guess_type(current_path)[0].split('/')[0]
    return mime_type in filters


class FileUploader:
    def __init__(self, element=None):
        self.element = element
        self.storage_client = storage.storage_factory(element=element)

    def upload_multiple_files(self, fileable, files=None, commit_sha1=None, prefix=None):
        if files is not None:
            chunks = FileUploader.chunk_file_list(files, math.ceil(len(files) / 1000))

            progress = Progressbar(total=len(files))

            progress_bar_lock = threading.Lock()

            files_queue = Queue(5000)
            progress_queue = Queue(5000)

            work = threading.Event()
            work.set()

            files_collector_thread = threading.Thread(
                target=FileUploader.files_collector,
                args=(files_queue, progress, progress_bar_lock, fileable, commit_sha1, chunks, prefix)
            )
            files_collector_thread.start()

            files_reporter_thread = threading.Thread(
                target=FileUploader.files_reporter,
                args=(progress_queue, progress, progress_bar_lock, fileable, commit_sha1)
            )
            files_reporter_thread.start()

            thread_pool = []
            for i in range(len(chunks)):
                t = threading.Thread(
                    target=FileUploader.worker,
                    args=(work, files_queue, progress_queue, self.storage_client)
                )
                t.start()
                thread_pool.append(t)

            files_collector_thread.join()
            files_reporter_thread.join()

            work.clear()

            for t in thread_pool:
                t.join()

    @staticmethod
    def files_collector(file_queue, progress, progress_bar_lock, fileable, commit_sha1, chunks, prefix):
        for chunk in chunks:
            try:
                blob_vs = FileUploader.create_blob_versions(fileable, commit_sha1, chunk, prefix)
                for k, v in blob_vs["files"].items():
                    try:
                        file_queue.put({
                            "full_path": k,
                            "source": v["local_path"],
                            "target": v["path"],
                            "bv_id": str(v["bv_id"])
                        })
                    except Exception:
                        with progress_bar_lock:
                            progress.progress += 1
                        print("Failed to upload {}".format(k))

                with progress_bar_lock:
                    progress.progress += len(chunk) - len(blob_vs["files"])

            except Exception as e:
                with progress_bar_lock:
                    progress.progress += len(chunk)
                print(e)

    @staticmethod
    def files_reporter(progress_queue, progress, progress_bar_lock, fileable, commit_sha1):
        uploaded_files = []

        while progress.progress < progress.total:
            try:
                file = progress_queue.get_nowait()

                with progress_bar_lock:
                    progress.progress += 1

                if file is not None:
                    uploaded_files.append(file["bv_id"])

                    if len(uploaded_files) >= 1000 or progress.progress >= progress.total:
                        FileUploader.save_blob_versions_to_commit(fileable, uploaded_files, commit_sha1)
                        uploaded_files = []
            except Exception:
                time.sleep(1)

        progress.finish()

    @staticmethod
    def worker(work, files_queue, progress_queue, storage_client):
        while work.is_set():
            try:
                file = files_queue.get_nowait()
                try:
                    storage_client.upload_single_file(file["source"], file["target"])
                    progress_queue.put(file)
                except Exception as e:
                    progress_queue.put(None)
                    print("Failed to upload {}".format(file["full_path"]))
                    print(e)

            except Exception as e:
                print(e)
                time.sleep(1)

    @staticmethod
    def chunk_file_list(file_list, num_chunks):
        chunks = []
        for i in range(0, num_chunks):
            chunks.append(file_list[i::num_chunks])
        return chunks

    @staticmethod
    def get_recursive_tree(file_path, is_absolute, is_dir, depth=False):
        trees = []
        if is_absolute:
            if depth:
                while str(file_path) != Path(file_path).anchor and str(file_path) != ".":
                    tmp_path = Path(file_path)
                    tmp_parent = tmp_path.parent
                    trees.append("{}/".format(str(tmp_parent)))
                    file_path = tmp_parent
            else:
                if is_dir and not file_path.endswith("/"):
                    file_path = "{}/".format(file_path)
                trees.append(file_path)
        else:
            while str(file_path) != Path(file_path).anchor and str(file_path) != ".":
                tmp_path = Path(file_path)
                tmp_parent = tmp_path.parent
                trees.append("{}/".format(str(tmp_parent)))
                file_path = tmp_parent
        return trees

    @staticmethod
    def get_files_and_dirs_recursive(root_dir=".", regex="*", filters=None):
        full_paths = [url_join(FileUploader.get_relative_path(root_dir), "/")]
        for root, dirs, files in walk(root_dir, topdown=False):
            for name in files:
                full_paths.append(FileUploader.get_relative_path(path.join(root, name)))
            for name in dirs:
                full_paths.append(url_join(FileUploader.get_relative_path(path.join(root, name)), "/"))

        if filters:
            full_paths = [item for item in files if filter_by_mimetypes(filters, item)]

        return fnmatch.filter(full_paths, regex)

    @staticmethod
    def get_relative_path(full_path):
        if full_path.startswith("./"):
            full_path = full_path[2:]
        return full_path

    @staticmethod
    def sha1(full_path):
        BUF_SIZE = 65536
        sha1 = hashlib.sha1()

        with open(full_path, 'rb') as f:
            while True:
                data = f.read(BUF_SIZE)
                if not data:
                    break
                sha1.update(data)

        return sha1.hexdigest()

    @staticmethod
    def create_blob_versions(fileable, commit_sha1, full_paths, prefix):
        owner = fileable.owner
        files = {}
        for full_path in full_paths:
            full_path = full_path.replace("./", "")
            is_file = os.path.isfile(full_path)
            is_directory = os.path.isdir(full_path)
            is_absolute = os.path.isabs(full_path)
            if is_absolute:
                relative_path = os.path.basename(full_path)
            else:
                relative_path = full_path

            if is_file:
                prefix_full_path = os.path.join(prefix, relative_path) if prefix else relative_path
                content_type = mimetypes.guess_type(full_path, strict=True)
                files.update({
                    prefix_full_path: {
                        "absolute_path": "{}/".format(os.path.dirname(os.path.abspath(full_path))),
                        "local_path": full_path,
                        "content_type": content_type[0] or "plain/text",
                        "file_name": os.path.basename(full_path),
                        "file_size": os.path.getsize(full_path),
                        "relative_path": prefix_full_path,
                        "sha1": FileUploader.sha1(full_path)
                    }
                })
            elif is_directory:
                files.update({
                    full_path: {
                        "absolute_path": "{}/".format(url_join(os.getcwd(), full_path)),
                        "relative_path": full_path
                    }
                })

        resp = apis_helper.post_v2(url_join(owner, "files", "create_blob_versions"), data={
            "fileable_type": "Project",
            "fileable_slug": fileable.slug,
            "commit": commit_sha1,
            "files": json.dumps(files)
        })

        return resp.json().get("result")

    @staticmethod
    def save_blob_versions_to_commit(fileable, blob_ids, commit_sha1):
        owner = fileable.owner
        resp = apis_helper.post_v2(url_join(owner, "files", "save_blob_versions_to_commit"), data={
            "fileable_type": "Project",
            "fileable_slug": fileable.slug,
            "commit": commit_sha1,
            "blob_ids": blob_ids
        })

        return resp.json()



import re


def __clean_start_end_slashes(s):
    if s.endswith("/"): s = s[:-1]
    if s.startswith("/"): s = s[1:]
    return s


def url_join(path, *paths):
    all_paths = [path] + list(paths)
    joined_url = "/".join(map(__clean_start_end_slashes, filter(lambda x: x, all_paths)))
    return re.sub(r'[\/]{2,}', "/", joined_url).replace("http:/", "http://").replace("https:/", "https://")



import os
import json
import mimetypes

from cnvrg.helpers.apis_helper import post as apis_post, get as apis_get, credentials
from cnvrg.modules.cnvrg_files import CnvrgFiles
from cnvrg.modules.errors import CnvrgError, APINotFound
from cnvrg.helpers.url_builder_helper import url_join
from cnvrg.helpers.error_catcher import suppress_exception
from cnvrg.modules.storage.file_uploader import FileUploader, filter_by_mimetypes

import cnvrg.helpers.config_helper as config_helper
import cnvrg.helpers.env_helper as env_helper
import cnvrg.helpers.args_helper as args_helper
import cnvrg.helpers.param_build_helper as param_build_helper
import cnvrg.helpers.apis_helper as apis_helper
import cnvrg.helpers.spawn_helper as spawn_helper


class Project(CnvrgFiles):
    def __init__(self, project=None, project_url=None, working_dir=None):
        if project_url:
            owner_slug, project_slug = Project.get_owner_and_project_from_url(project_url)
        else:
            owner_slug, project_slug = param_build_helper.parse_params(project, param_build_helper.PROJECT, working_dir=working_dir)
        self.__owner = owner_slug or credentials.owner
        self.__project = project_slug

        in_dir = config_helper.is_in_dir(config_helper.CONFIG_TYPE_PROJECT, project_slug, working_dir)
        super(Project, self).__init__(in_dir=in_dir)
        if in_dir:
            self._set_working_dir(config_helper.find_config_dir(path=working_dir))
            self.in_dir = True
        elif not self.__project or not self.__owner:
            raise CnvrgError("Cant init project without params and outside project directory")

    def __fetch_config(self):
        resp = apis_get(self.get_base_url())
        return {"slug": self.slug, "owner": self.owner, **json.loads(resp.get("result"))}

    def is_git(self):
        resp = apis_get(url_join(self.get_base_url(), 'get_project'))
        if not resp.get("result"):
            return False
        result = json.loads(resp.get("result"))
        return result.get("git")

    def put_files(self, paths=None, pattern="*", message="SDK Commit", depth=False, payload=None, prefix=None, filters=None, tag_images=False):
        if not paths:
            raise CnvrgError("Please send files or path")

        full_paths = []

        for p in paths:
            is_absolute = os.path.isabs(p)
            is_dir = os.path.isdir(p)
            is_file = os.path.isfile(p)

            if is_dir:
                full_paths += FileUploader.get_files_and_dirs_recursive(root_dir=p, regex=pattern, filters=filters)
            elif is_file and filter_by_mimetypes(filters, p):
                trees = FileUploader.get_recursive_tree(p, is_absolute, is_dir, depth)
                full_paths += trees
                if not is_absolute:
                    if is_dir:
                        file_path = "{}/".format(p)
                    else:
                        file_path = p
                    full_paths.append(file_path)

                if is_absolute and depth:
                    if is_dir:
                        file_path = "{}/".format(p)
                    else:
                        file_path = p
                    full_paths.append(file_path)
            else:
                print("{} not exists skipping it".format(p))
                try:
                    paths.remove(p)
                except ValueError:
                    pass

        full_paths_set = set(full_paths) - {"./", ".", "/", "//"}
        unique_full_paths = list(full_paths_set)

        if not unique_full_paths:
            print("Nothing to upload\nAborting")
            exit(0)

        fu = FileUploader(self)
        commit_sha1 = self.start_commit_new(message=message, payload=payload, workflow_slug=payload['workflow_slug'])
        fu.upload_multiple_files(self, files=unique_full_paths, commit_sha1=commit_sha1, prefix=prefix)
        resp = self.end_commit_new(commit_sha1=commit_sha1, tag_images=tag_images, workflow_slug=payload['workflow_slug'])
        if not resp.get("data"):
            raise CnvrgError("Couldn’t create commit")
        else:
            print("Successfully created commit {}".format(resp.get("data").get("attributes").get("sha1")))

    def start_commit_new(self, message="SDK Commit", payload=None, workflow_slug=''):
        if payload is None:
            payload = {}

        resp = apis_helper.post_v2(url_join(self.owner, 'projects', self.get_project_name(), 'workflows', workflow_slug, "commits"), data={
            "job_slug": workflow_slug,
            "message": message,
            "source": "sdk-old",
            **payload
        })
        if resp.status_code == 404:
            raise APINotFound("API Not Found")
        return resp.json().get('data').get('attributes').get('sha1')

    def end_commit_new(self, commit_sha1, tag_images, workflow_slug=''):
        resp = apis_helper.put_v2(url_join(self.owner, 'projects', self.get_project_name(), 'workflows', workflow_slug, "commits", commit_sha1, "end_commit"), data={
            "tag_images": tag_images,
        })

        return resp.json()

    @staticmethod
    @suppress_exception
    def list(owner=None):
        owner_slug = owner or credentials.owner
        projects = []
        resp = apis_helper.get_v2(url_join(owner_slug, "projects"))
        for proj in resp.json()["data"]:
            slug = proj["attributes"]["slug"]
            projects.append(Project("{}/{}".format(owner_slug, slug)))
        return projects

    @staticmethod
    @suppress_exception
    def create(title, owner=None):
        owner_slug = owner or credentials.owner
        resp = apis_helper.post_v2(url_join(owner_slug, "projects"), data={"title": title})
        status = resp.status_code
        if status == 200:
            slug = resp.json()["data"]["attributes"]["slug"]
            print("Project {} created successfully".format(slug))
            return Project(slug)
        else:
            print("Could not create project")

    @staticmethod
    def factory(owner, project):
        if isinstance(project, str):
            return Project(url_join(owner, project))
        elif isinstance(project, Project):
            return project

    @property
    def owner(self):
        return self.__owner

    @property
    def slug(self):
        return self.__project

    @property
    def git(self):
        config = self.get_config()
        if config:
            return config.get(":git") or False
        return self.__fetch_config().get("git")

    def get_git_commit(self):
        return spawn_helper.run_and_get_output("git rev-parse --verify HEAD")

    def get_git_branch(self):
        return spawn_helper.run_and_get_output("git rev-parse --abbrev-ref HEAD")

    def get_base_url(self, api="v1"):
        if api == "v1":
            return "users/{owner}/projects/{project}".format(owner=self.__owner, project=self.__project)
        else:
            return "{owner}/projects/{project}".format(owner=self.__owner, project=self.__project)

    def _default_config(self):
        return {
            "project": self.__project,
            "owner": self.__owner,
            "commit": None
        }

    def web_url(self):
        return url_join(credentials.web_url(), self.__owner, 'projects', self.__project)

    def get_project_name(self):
        return self.__project

    def get_output_dir(self):
        return self.get_working_dir()

    def run_task(self, cmd, **kwargs):
        #if "grid" in kwargs: kwargs["grid"] = hyper_search_helper.load_hyper_search(kwargs["grid"])
        kwargs = {**env_helper.get_origin_job(), **kwargs}
        resp = apis_post(url_join(self.get_base_url(), "experiments"), data={"cmd": cmd, **kwargs})
        hyper_search = resp.get("hyper_search")
        if not hyper_search:
            error_msg = resp.get("error") or "Can't run experiment"
            raise CnvrgError(error_msg)
        return hyper_search

    def sync(self, get_output_by_regex=".*", **options):
        command = "cnvrg sync {args}".format(args=args_helper.args_to_string(options))
        return spawn_helper.run_sync(command, print_output=True, get_output_by_regex=get_output_by_regex)



from cnvrg.modules.base_module import CnvrgBase
from cnvrg.modules.project import Project
import cnvrg.helpers.apis_helper as apis_helper
import cnvrg.helpers.string_helper as string_helper
import os
import cnvrg.helpers.env_helper as env_helper
import cnvrg.helpers.config_helper as config_helper
import cnvrg.helpers.time_helper as time_helper
from cnvrg.modules.errors import UserError, CnvrgError
from cnvrg.helpers.url_builder_helper import url_join
import cnvrg.helpers.files_helper as files_helper
from cnvrg.modules.storage import storage_factory


LOGS_TYPE_OUTPUT = "output"
LOGS_TYPE_ERROR = "error"
LOGS_TYPE_INFO = "info"
LOGS_TYPE_WARNING = "warning"
MAX_LOGS_PER_SEND = 400

class CnvrgJob(CnvrgBase):
    def __init__(self, slug=None, job_type=None, project=None):
        self.job_slug = slug or env_helper.CURRENT_JOB_ID
        self.job_type = job_type or env_helper.CURRENT_JOB_TYPE
        if not project:
            if not config_helper.config_type() == config_helper.CONFIG_TYPE_PROJECT:
                raise UserError(
                    "Cant create an experiment without a project, please pass a project or cd into cnvrg project dir")
            project = Project()
        self.project = project
        self.job_project_slug = project.slug
        self.job_owner_slug = project.owner


    def _base_job_url(self):
        return url_join(
            #### hackish :D
            self.project.get_base_url(),"jobs", string_helper.to_snake_case(self.job_type), self.job_slug
        )


    def send_util(self, utilization):
        apis_helper.post(url_join(self._base_job_url(), "utilization"), data=[utilization])


    def log(self, logs, log_type=LOGS_TYPE_OUTPUT):
        if isinstance(logs, str): logs = [logs]
        for i in range(0, len(logs), env_helper.MAX_LOGS_PER_SEND):
            apis_helper.post(url_join(self._base_job_url(), "log"), data={"logs": logs[i:i + env_helper.MAX_LOGS_PER_SEND], "log_level": log_type, "timestamp": time_helper.now_as_string()})


    def log_artifacts(self, files: list, target="output"):
        file_dict = {}
        for file in files:
            props = files_helper.get_file_props(file)
            ### we are uploading the file to target/file_name.
            target_path = url_join(target, props["file_name"])
            props["relative_path"] = target_path
            file_dict[target_path] = {"local_path": file, "props": props}
        resp = apis_helper.post(url_join(self._base_job_url(), "log_artifacts"),
                                data={"files": [file.get("props") for file in file_dict.values()]})
        commit_sha1 = resp.get("commit")
        if not commit_sha1: raise CnvrgError("Cant upload files")
        storage = storage_factory(self.project)
        files = resp.get("files").get("files")
        std_files = []
        for k,v in files.items():
            std_files.append({
                "local_path": file_dict.get(k).get("local_path"),
                "storage_path": v.get("path"),
                "decrypted": True
            })
        storage.upload_files(std_files)
        resp = apis_helper.post(url_join(self._base_job_url(), "log_artifacts_end"),
                                data={"commit": resp.get("commit")})
        return commit_sha1




    @staticmethod
    def current_job_sync_args(output_dir=None, **kwargs):
        options = {}
        options["job"] = env_helper.CURRENT_JOB_ID
        options["job_type"] = env_helper.CURRENT_JOB_TYPE
        options["in_exp"] = "true"
        output_dir = output_dir or env_helper.CNVRG_OUTPUT_DIR
        if output_dir: options["output_dir"] = output_dir
        options = {**options, **kwargs}
        return options





import os

from typing import Dict
from google.cloud import storage
from cnvrg.modules.cnvrg_files import CnvrgFiles
from cnvrg.helpers.apis_helper import download_file
from cnvrg.modules.storage.base_storage import BaseStorage


class GCPStorage(BaseStorage):
    def __init__(self, element: CnvrgFiles, working_dir: str, storage_resp: Dict):
        super().__init__(element, working_dir, storage_resp.get("sts"))

        try:
            os.remove(self.sts_local_file)
        except Exception:
            pass

        del storage_resp["sts"]

        self.key_file = os.path.join(os.path.expanduser("~"), ".cnvrg", ".gcp_cred.json")

        props = self.decrypt_dict(storage_resp, keys=["credentials"])

        download_file(props["credentials"], self.key_file)

        self.client = self._get_client()
        self.bucket = self.client.bucket("cnvrg-storage")

    def upload_single_file(self, file, target):
        try:
            bucket = self.bucket
            blob = bucket.blob(target)
            blob.upload_from_filename(file)
        except Exception as e:
            print(e)

    def download_single_file(self, file, target):
        pass

    def _get_client(self):
        return storage.Client.from_service_account_json(self.key_file)



import os
import yaml

from pathlib import Path

import cnvrg.helpers.env_helper as env_helper

CONFIG_TYPE_PROJECT = "PROJECT"
CONFIG_TYPE_DATASET = "DATASET"

def find_config_dir(path=None):
    currdir = path or os.curdir
    real_path = os.path.realpath(currdir)
    home_path = os.path.expanduser("~")
    while real_path != home_path and real_path != Path(home_path).anchor and real_path != ".":
        if os.path.exists(os.path.join(currdir, '.cnvrg', 'config.yml')):
            return currdir
        tmp_path = Path(real_path)
        currdir = tmp_path.parent
        real_path = os.path.realpath(currdir)
    return False


def config_path(path=None):
    config_dir = find_config_dir(path=path)
    if not config_dir: return
    return os.path.join(config_dir, ".cnvrg", "config.yml")


def load_config(path=None):
    if not path: return {}
    with open(config_path(path), 'r') as f:
        return yaml.safe_load(f)


def config_type(path=None):
    if not find_config_dir(path):
        return None
    config = load_config(path=path)
    if "project" in config or ":project_name" in config:
        return CONFIG_TYPE_PROJECT
    return CONFIG_TYPE_DATASET


def save_config(config, path=None):
    config_dir = find_config_dir(path=path)
    config_dir = os.path.join(config_dir or path, ".cnvrg")
    os.makedirs(config_dir, exist_ok=True)
    config_path = os.path.join(config_dir, "config.yml")
    with open(config_path, "w+") as f:
        yaml.dump(config, f)

def idx_path(path=None):
    config_dir = find_config_dir(path=path)
    if not config_dir: return
    return os.path.join(config_dir, ".cnvrg", "idx.yml")

def load_idx(path=None):
    if not path: return {}

    with open(idx_path(path), 'r') as f:
        return yaml.safe_load(f)

def is_in_dir(dir_type, dir_element, working_dir):
    working_dir = working_dir or os.curdir
    if not find_config_dir(working_dir): return False
    if not config_type(working_dir) == dir_type: return False
    config = load_config(working_dir)
    return (config.get(":project_slug") or config.get("project") or config.get(":dataset_slug") or config.get("dataset_name") or config.get("dataset")) == dir_element


def get_element_slug(working_dir=None):
    config = load_config(find_config_dir(working_dir))
    return env_helper.CURRENT_PROJECT_SLUG or config.get(":project_slug") or config.get("project_slug") or config.get("project") or config.get(":dataset_slug") or config.get("dataset_slug") or config.get("dataset")


def get_element_owner(working_dir=None):
    config = load_config(find_config_dir(working_dir))
    return env_helper.CURRENT_ORGANIZATION_SLUG or config.get(":owner") or config.get("owner")

def general_config_path():
    return os.path.expanduser("~/.cnvrg/config.yml")

def load_generl_config():
    config_file = general_config_path()
    if not os.path.isfile(config_file):
        return ""
    with open(config_file, 'r') as f:
        return yaml.safe_load(f)
